# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rangeindex']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'rangeindex',
    'version': '0.8.2',
    'description': '',
    'long_description': '# RangeIndex\n\n[![tests Actions Status](https://github.com/manimino/rangeindex/workflows/tests/badge.svg)](https://github.com/manimino/rangeindex/actions)\n\nContainer for finding Python objects by `<`, `<=`, `==`, `>=`, `>` on their attributes.\n\n`pip install rangeindex`\n\n### Usage\n```\nfrom rangeindex import RangeIndex\nri = RangeIndex(list_of_objects, on={\'size\': int, \'brightness\': float})\nmatching_objects = ri.find(\'size >= 1000 and brightness > 0.5\')\n```\n\nYou can `add()`, `add_many()`, `update()`, and `remove()` items from a RangeIndex.\n\n### How it works\n\nWhen you do: `RangeIndex(list_of_objects, on={\'size\': int, \'brightness\': float})`\n\nA table is created with 3 columns:\n - size\n - brightness\n - Python object reference\n\nOn `find()`, a query will run to find the matching objects.\n\n### Example\n\nYou have a million cat photos. Find big, bright pictures of Tiger The Cat.\n\n```\nimport random\nimport time\nfrom rangeindex import RangeIndex\n\n\nclass CatPhoto:\n    def __init__(self):\n        self.width = random.choice(range(200, 2000))\n        self.height = random.choice(range(200, 2000))\n        self.brightness = random.random() * 10\n        self.name = random.choice(["Luna", "Willow", "Elvis", "Nacho", "Tiger"])\n        self.image_data = "Y2Ugbidlc3QgcGFzIHVuZSBjaGF0dGU="\n\n\n# Make a million\nphotos = [CatPhoto() for _ in range(10 ** 6)]\n\n# Build RangeIndex\nri = RangeIndex(\n    photos,\n    on={"height": int, "width": int, "brightness": float, "name": str},\n    engine="sqlite",\n    table_index=[("width", "height", "brightness")],\n)\n\n# Find matches\nimport time\n\nt0 = time.time()\nmatches = ri.find(\n    "height > 1900 and width >= 1900 and brightness >= 9 and name=\'Tiger\'"\n)\nprint(\'Found\', len(matches), \'matches.\')\n```\n\nIn this case, RangeIndex `find()` is more than 10x faster than the equivalent Python expression:\n\n`matches = [p for p in photos if p.height >= 1900 and p.width >= 1900 and p.brightness >= 9 and p.name=\'Tiger\']`\n\n### Engine Comparison\n\nRangeIndex has two engines available, `sqlite` and `pandas`. The default is `sqlite`.\n\nIf your queries typically return just a few results, use `engine=\'sqlite\'`. But if you\'re doing full table \nscans often, `engine=\'pandas\'` will be faster. \n\n#### Time Comparison\n\n|                 | Baseline | Sqlite | Pandas |\n|-----------------|----------|--------|--------|\n| Find 1 item     | 0.9s     | 0.2ms  | 43.1ms |\n| Find 10 items   | 0.9s     | 0.7ms  | 44.9ms |\n| Find 100 items  | 1.0s     | 1.9ms  | 43.8ms |\n| Find 1K items   | 1.0s     | 6.7ms  | 43.9ms |\n| Find 10K items  | 1.1s     | 27.2ms | 47.6ms |\n| Find 100K items | 1.2s     | 0.19s  | 88.3ms |\n| Find 1M items   | 1.7s     | 1.37s  | 0.24s  |\n| Find 10M items  | 2.9s     | 10.6s  | 0.45s  |\n\nThis is a benchmark on random range queries against a dataset of 10 million (10^7) objects indexed on two numeric \nfields. `Baseline` is a Python list comprehension.\n\n#### Graph\n\n![Benchmark: sqlite does well on small queries, other engines do better on large queries.](perf/benchmark.png)\n\nThis is the same data in graph form, showing relative speedup. Each line is divided by `baseline`. \nNote that both axis labels are powers of 10. So `10^3` on the Y-axis indicates a 1000X speedup.\n\n## Methods\n\n### Init\n\n```\nRangeIndex(\n        objs: Optional[Iterable[Any]] = None,\n        on: Optional[Dict[str, Any]] = None,\n        engine: str = SQLITE,\n        **kwargs\n)\n```\n\nCreates a RangeIndex.\n\n`objs` is optional. It can be any collection of `class`, `dataclass`, or `namedtuple` objects.\n\n`on` is required. It specifies the attributes and types to index. \nThe allowed types are `float`, `int`, `bool`, and `str`.\n\n`engine` is either `\'sqlite\'` or `\'pandas\'`, defaults to `sqlite`. Pandas is an optional dependency, so \n`pip install pandas` if you want to use the `engine=\'pandas\'`.\n\nIf the engine is sqlite, you may optionally specify `table_index`. This controls the table index that SQLite uses when \nperforming queries. If unspecified, a single-column index is made on each\nattribute. Example: `table_index=[(\'a\', \'b\', \'c\'), (\'d\')]` will create a multi-column index on `(a, b, c)` and a \nsingle-column index on `d`. Multi-column indexes will often speed up `find()` operations; see \n[SQLite documentation](https://www.sqlite.org/queryplanner.html).\n\n### add(), add_many()\n\n```\nadd(obj:Any)\nadd_many(objs:Iterable[Any])\n```\n\nYou can add a single object with `add()`. If you have many objects, it is much faster to `add_many()` than it is to\ncall `add()` on each.\n\nIf an added object is missing an attribute, the object will still be added. The missing attribute will be given a \nnull value in the index.\n\n### find()\n\n`find(where: Optional[str]) -> List` finds objects matching the query string in `where`.\n\nExamples: \n - `ri.find(\'b == True and string == "okay"\')`\n - `ri.find(\'(x == 0 and y >= 1000.0) or x == 9\')`\n\nIf `where` is unspecified, all objects in the RangeIndex are returned. \n\nThe syntax of `where` is nearly identical between pandas and sqlite. The only difference is in matching null \nvalues. \n - In sqlite, use `find(\'x is null\')` / `find(\'x is not null\')`. \n - In pandas, use `find(\'x != x\')` to match nulls, or `find(\'x == x\')` for non-nulls. \n\nConsult the syntax for [SQLite queries](https://www.sqlite.org/lang_select.html) or \n[pandas queries](https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.query.html) as needed.\n\n### update()\n\n`update(self, obj: Any, updates: Dict[str, Any])` updates attributes of a single object in the index.\n\n`updates` is a dict containing the new values for each changed attribute, e.g. `{\'x\': 5.5, \'b\': True}`.\n\nIf you change an indexed object\'s attributes without calling `update()`, the RangeIndex will be out of sync and\nreturn inaccurate results. \n\n`update()` will changes both the value in the RangeIndex table and the object\'s value.\n\nUpdate is fast (less than 1 ms), it\'s O(log n) in both sqlite and pandas.\n\n### remove()\n\n`remove(self, obj: Any)` removes an object. \n\nRemove is fast (less than 1ms) in SQLite but slower (tens of ms) in Pandas. \nThis is because removing an item requires rebuilding arrays there.\n\n### Container methods\n\nYou can do the container things:\n - Length: `len(ri)`\n - Contains: `obj in ri`\n - Iteration: `for obj in ri: ...`\n',
    'author': 'Theo Walker',
    'author_email': 'theo.ca.walker@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/manimino/rangeindex/',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
